/*-
 * Copyright (c) 1998 Brian Somers <brian@Awfulhak.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $FreeBSD: src/usr.sbin/ppp/iface.c,v 1.15 2000/06/19 21:43:48 brian Exp $
 */

#define _GNU_SOURCE
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <sys/select.h>
#include <net/if.h>
#ifndef __GNU__
#include <net/if_dl.h>
#include <net/route.h>
#endif
#include <arpa/inet.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <sys/un.h>

#include <errno.h>
#include <string.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#ifndef __GNU__
#include <sys/sysctl.h>
#endif
#include <termios.h>
#include <unistd.h>

#include <hurd.h>
#include <hurd/socket.h>

#include "pfinet.h"

#include "layer.h"
#include "defs.h"
#include "command.h"
#include "mbuf.h"
#include "log.h"
#include "id.h"
#include "timer.h"
#include "fsm.h"
#include "iplist.h"
#include "lqr.h"
#include "hdlc.h"
#include "throughput.h"
#include "slcompress.h"
#include "descriptor.h"
#include "ipcp.h"
#include "filter.h"
#include "lcp.h"
#include "ccp.h"
#include "link.h"
#include "mp.h"
#ifndef NORADIUS
#include "radius.h"
#endif
#include "bundle.h"
#include "prompt.h"
#include "iface.h"


static int
bitsinmask(struct in_addr mask)
{
  u_int32_t bitmask, maskaddr;
  int bits;

  bitmask = 0xffffffff;
  maskaddr = ntohl(mask.s_addr);
  for (bits = 32; bits >= 0; bits--) {
    if (maskaddr == bitmask)
      break;
    bitmask &= ~(1 << (32 - bits));
  }

  return bits;
}


#if 1
#define iface_addr_Zap(name, addr) ({ 0; })
#else
/* int iface_addr_Zap(const char *name, struct iface_addr *addr) */
#define iface_addr_Zap(name, addr) ({					\
  int s;								\
  struct ifreq ifr;							\
  struct sockaddr_in *me, *peer;					\
  int ret = 1;								\
									\
  s = ID0socket(AF_INET, SOCK_DGRAM, 0);				\
  if (s < 0) {								\
    log_Printf(LogERROR, __FUNCTION__ ": socket(): %s\n", strerror(errno));\
    ret = 0;								\
  } else {								\
    memset(&ifr, '\0', sizeof ifr);					\
    strncpy (ifr.ifr_name, (name), sizeof ifr.ifr_name - 1);		\
    me = (struct sockaddr_in *)&ifr.ifr_addr;				\
    peer = (struct sockaddr_in *)&ifr.ifr_broadaddr;			\
    me->sin_family = peer->sin_family = AF_INET;			\
    me->sin_len = peer->sin_len = sizeof(struct sockaddr_in);		\
    me->sin_addr = (addr)->ifa;						\
    peer->sin_addr = (addr)->brd;					\
    if (ID0ioctl(s, SIOCDIFADDR, &ifr) < 0) {				\
      log_Printf(LogWARN, __FUNCTION__ ": ioctl(SIOCDIFADDR, %s): %s\n",\
                 inet_ntoa((addr)->ifa), strerror(errno));		\
      ret = 0;								\
    }									\
    close (s);								\
  }									\
									\
  ret; })
#endif

struct iface *
iface_Create(const char *name)
{
  int s;
  struct ifreq ifr;
  struct iface *iface = NULL;

  s = ID0socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    log_Printf(LogERROR, __FUNCTION__ ": socket(): %s.\n", strerror(errno));
    goto lose;
  }

  iface = malloc(sizeof (struct iface));
  if (! iface) {
    log_Printf(LogERROR, __FUNCTION__ ": malloc: %s.\n", strerror(errno));
    goto lose_socket;
  }

  iface->in_addr = malloc(sizeof (struct iface_addr));
  if (! iface->in_addr) {
    log_Printf(LogERROR, __FUNCTION__ ": malloc: %s.\n", strerror(errno));
    goto lose_iface;
  }

  memset(&ifr, '\0', sizeof ifr);
  strncpy (ifr.ifr_name, name, IFNAMSIZ);
  ifr.ifr_name[IFNAMSIZ] = '\0';
  if (ioctl(s, SIOCGIFFLAGS, &ifr) < 0) {
    log_Printf(LogERROR, __FUNCTION__ ": iioctl_siocgifflags: %s.\n",
	       strerror(errno));
    goto lose_addr;
  }

  iface->name = strdup(name);
  iface->flags = ifr.ifr_flags;
  iface->index = ifr.ifr_ifindex;
  iface->in_addrs = 0;

  if (ID0ioctl (s, SIOCGIFADDR, &ifr) < 0) {
    log_Printf(LogERROR, __FUNCTION__ ": getting address: %s.\n",
	       strerror(errno));
    goto lose_name;
  } else {
    iface->in_addr->ifa = ((struct sockaddr_in *)&ifr.ifr_addr)->sin_addr;
    iface->in_addrs ++;
  }

  if (ID0ioctl (s, SIOCGIFDSTADDR, &ifr) < 0)
    iface->in_addr->brd.s_addr = INADDR_ANY;
  else
    iface->in_addr->brd = ((struct sockaddr_in *)&ifr.ifr_broadaddr)->sin_addr;

  if (ID0ioctl (s, SIOCGIFNETMASK, &ifr) < 0)
    iface->in_addr->mask.s_addr = INADDR_ANY;
  else
    iface->in_addr->mask = ((struct sockaddr_in *)&ifr.ifr_netmask)->sin_addr;

  iface->in_addr->bits = bitsinmask(iface->in_addr->mask);

  close (s);
	
  return iface;

lose_name:
  free (iface->name);
lose_addr:
  free (iface->in_addr);
lose_iface:
  free (iface);
lose_socket:
  close (s);
lose:
  return NULL;
}

void
iface_inClear(struct iface *iface, int how)
{
  if (iface->in_addrs) {
    if (iface_addr_Zap(iface->name, &iface->in_addr[0]))
      iface->in_addrs = 0;
  }
}

int
iface_inAdd(struct iface *iface, struct in_addr ifa, struct in_addr mask,
            struct in_addr brd, int how)
{
  int s;
  struct ifreq ifr;
  struct sockaddr_in *addr = (struct sockaddr_in *) &ifr.ifr_addr;

  if (! iface->in_addr) {
    iface->in_addr = malloc(sizeof (struct iface_addr));
    if (! iface->in_addr) {
      log_Printf(LogERROR, __FUNCTION__ ": malloc: %s\n", strerror(errno));
      return 0;
    }
  }

  s = ID0socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    log_Printf(LogERROR, __FUNCTION__ ": socket(): %s\n", strerror(errno));
    return 0;
  }

  memset(&ifr, '\0', sizeof ifr);
  strncpy (ifr.ifr_name, iface->name, IFNAMSIZ - 1);

  if (log_IsKept(LogDEBUG)) {
    char buf[16];

    strncpy(buf, inet_ntoa(brd), sizeof buf-1);
    buf[sizeof buf - 1] = '\0';
    log_Printf(LogDEBUG, "Add %s -> %s\n", inet_ntoa(ifa), buf);
  }

  addr->sin_addr = ifa;
  addr->sin_len = sizeof (struct sockaddr_in);
  addr->sin_family = AF_INET;
  if ((ID0ioctl(s, SIOCSIFADDR, &ifr) < 0)
      && (brd.s_addr != INADDR_BROADCAST || errno != EEXIST)) {
    log_Printf(LogERROR, __FUNCTION__ ": ioctl(SIOCSIFADDR): %s: %s\n",
               inet_ntoa(ifa), strerror(errno));
    ID0ioctl(s, SIOCDIFADDR, &ifr); /* Be really sure that nothing got in.  */
    close(s);
    return 0;
  }

  addr->sin_addr = brd;
  addr->sin_len = sizeof (struct sockaddr_in);
  addr->sin_family = AF_INET;
  if (ID0ioctl(s, SIOCSIFDSTADDR, &ifr) < 0) {
    log_Printf(LogERROR, __FUNCTION__ ": ioctl(SIOCSIFDSTADDR): %s: %s\n",
               inet_ntoa(ifa), strerror(errno));
    ID0ioctl(s, SIOCDIFADDR, &ifr);
    close(s);
    return 0;
  }

  addr->sin_addr = mask;
  addr->sin_len = sizeof (struct sockaddr_in);
  addr->sin_family = AF_INET;
  if ((ID0ioctl(s, SIOCSIFNETMASK, &ifr) < 0)) {
    log_Printf(LogERROR, __FUNCTION__ ": ioctl(SIOCSIFNETMASK): %s: %s\n",
               inet_ntoa(mask), strerror(errno));
    ID0ioctl(s, SIOCDIFADDR, &ifr);
    close(s);
    return 0;
  }

  close(s);

  iface->in_addr[0].ifa = ifa;
  iface->in_addr[0].mask = mask;
  iface->in_addr[0].brd = brd;
  iface->in_addr[0].bits = bitsinmask(iface->in_addr[0].mask);

  return 1;
}

int
iface_inDelete(struct iface *iface, struct in_addr ip)
{
  if (iface->in_addr[0].ifa.s_addr == ip.s_addr) {
    if (iface_addr_Zap (iface->name, &iface->in_addr[0]))
      iface->in_addrs = 0;
    return 1;
  }

  return 0;
}

#define IFACE_ADDFLAGS 1
#define IFACE_DELFLAGS 2

static int
iface_ChangeFlags(struct iface *iface, int flags, int how)
{
  struct ifreq ifrq;
  int s;

  s = ID0socket(AF_INET, SOCK_DGRAM, 0);
  if (s < 0) {
    log_Printf(LogERROR, "iface_ChangeFlags: socket: %s\n", strerror(errno));
    return 0;
  }

  memset(&ifrq, '\0', sizeof ifrq);
  strncpy(ifrq.ifr_name, iface->name, sizeof ifrq.ifr_name - 1);
  ifrq.ifr_name[sizeof ifrq.ifr_name - 1] = '\0';
  if (ID0ioctl(s, SIOCGIFFLAGS, &ifrq) < 0) {
    log_Printf(LogERROR, "iface_ChangeFlags: ioctl(SIOCGIFFLAGS): %s\n",
	       strerror(errno));
    close(s);
    return 0;
  }

  if (how == IFACE_ADDFLAGS)
    ifrq.ifr_flags |= flags;
  else
    ifrq.ifr_flags &= ~flags;

  if (ID0ioctl(s, SIOCSIFFLAGS, &ifrq) < 0) {
    log_Printf(LogERROR, "iface_ChangeFlags: ioctl(SIOCSIFFLAGS): %s\n",
	       strerror(errno));
    close(s);
    return 0;
  }
  close(s);

  return 1;	/* Success */
}

int
iface_SetFlags(struct iface *iface, int flags)
{
  return iface_ChangeFlags(iface, flags, IFACE_ADDFLAGS);
}

int
iface_ClearFlags(struct iface *iface, int flags)
{
  return iface_ChangeFlags(iface, flags, IFACE_DELFLAGS);
}

void
iface_Destroy(struct iface *iface)
{
  /*
   * iface_Clear(iface, IFACE_CLEAR_ALL) must be called manually
   * if that's what the user wants.  It's better to leave the interface
   * allocated so that existing connections can continue to work.
   */

  if (iface != NULL) {
    free(iface->name);
    free(iface->in_addr);
    free(iface);
  }
}

#define if_entry(x) { IFF_##x, #x }

struct {
  int flag;
  const char *value;
} if_flags[] = {
  if_entry(UP),
  if_entry(BROADCAST),
  if_entry(DEBUG),
  if_entry(LOOPBACK),
  if_entry(POINTOPOINT),
  if_entry(RUNNING),
  if_entry(NOARP),
  if_entry(PROMISC),
  if_entry(ALLMULTI),

  /* These flags aren't available under the HURD */
  if_entry(MULTICAST),
  { 0, "???" }
};

int
iface_Show(struct cmdargs const *arg)
{
  struct iface *iface = arg->bundle->iface, *current;
  int f, flags;

  current = iface_Create(iface->name);
  flags = iface->flags = current->flags;
  iface_Destroy(current);

  prompt_Printf(arg->prompt, "%s (idx %d) <", iface->name, iface->index);
  for (f = 0; f < sizeof if_flags / sizeof if_flags[0]; f++)
    if ((if_flags[f].flag & flags) || (!if_flags[f].flag && flags)) {
      prompt_Printf(arg->prompt, "%s%s", flags == iface->flags ? "" : ",",
                    if_flags[f].value);
      flags &= ~if_flags[f].flag;
    }
  prompt_Printf(arg->prompt, "> has %d address%s:\n", iface->in_addrs,
                iface->in_addrs == 1 ? "" : "es");

  for (f = 0; f < iface->in_addrs; f++) {
    prompt_Printf(arg->prompt, "  %s", inet_ntoa(iface->in_addr[f].ifa));
    if (iface->in_addr[f].bits >= 0)
      prompt_Printf(arg->prompt, "/%d", iface->in_addr[f].bits);
    if (iface->flags & IFF_POINTOPOINT)
      prompt_Printf(arg->prompt, " -> %s", inet_ntoa(iface->in_addr[f].brd));
    else if (iface->flags & IFF_BROADCAST)
      prompt_Printf(arg->prompt, " broadcast %s",
                    inet_ntoa(iface->in_addr[f].brd));
    if (iface->in_addr[f].bits < 0)
      prompt_Printf(arg->prompt, " (mask %s)",
                    inet_ntoa(iface->in_addr[f].mask));
    prompt_Printf(arg->prompt, "\n");
  }

  return 0;
}
